#include "TargaImage.h"
#include <libtarga.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

static const unsigned char	BACKGROUND[3] = { 0, 0, 0 };


// Constructor. COPIES THE IMAGE DATA.
TargaImage::TargaImage(int w, int h, unsigned char *d)
{
    int i;

    width = w;
    height = h;
    data = new unsigned char[width * height * 4];

    for ( i = 0 ; i < width * height * 4 ; i++ )
	data[i] = d[i];
}


TargaImage*
TargaImage::Alpha_Image(void)
{
    unsigned char   *dest = new unsigned char[width * height * 4];
    TargaImage	    *result;
    int 	    i, j;

    if ( ! data )
	return NULL;

    for ( i = 0 ; i < height ; i++ )
    {
	int offset = i * width * 4;

	for ( j = 0 ; j < width ; j++ )
        {
	    dest[offset + j * 4] =
	    dest[offset + j * 4 + 1] =
	    dest[offset + j * 4 + 2] = data[offset + j * 4 + 3];
	    dest[offset + j * 4 + 3] = 255;
       }
    }

    result = new TargaImage(width, height, dest);

    delete[] dest;

    return result;
}


// Copy this into a new image, reversing the rows as it goes. A pointer
// to the new image is returned.
TargaImage*
TargaImage::Reverse_Rows(void)
{
    unsigned char   *dest = new unsigned char[width * height * 4];
    TargaImage	    *result;
    int 	    i, j;

    if ( ! data )
	return NULL;

    for ( i = 0 ; i < height ; i++ )
    {
	int in_offset = ( height - i - 1 ) * width * 4;
	int out_offset = i * width * 4;

	for ( j = 0 ; j < width ; j++ )
        {
	    dest[out_offset + j * 4] = data[in_offset + j * 4];
	    dest[out_offset + j * 4 + 1] = data[in_offset + j * 4 + 1];
	    dest[out_offset + j * 4 + 2] = data[in_offset + j * 4 + 2];
	    dest[out_offset + j * 4 + 3] = data[in_offset + j * 4 + 3];
       }
    }

    result = new TargaImage(width, height, dest);

    delete[] dest;

    return result;
}


// Save the image to a targa file. Returns 1 on success, 0 on failure.
int
TargaImage::Save_Image(char *filename)
{
    TargaImage	*out_image = Reverse_Rows();

    if ( ! out_image )
	return 0;

    if ( ! ( tga_write_raw(filename, width, height, out_image->data,
                           TGA_TRUECOLOR_32) ) )
    {
	fprintf(stderr, "TGA Save Error: %s\n",
	        tga_error_string(tga_get_last_error()));
	return 0;
    }

    delete out_image;

    return 1;
}


TargaImage*
TargaImage::Load_Image(char *filename)
{
    unsigned char   *temp_data;
    TargaImage	    *temp_image;
    TargaImage	    *result;
    int		    width, height;

    temp_data = (unsigned char*)tga_load(filename, &width, &height,
                                         TGA_TRUECOLOR_32);
    if ( ! temp_data )
    {
	fprintf(stderr, "TGA Error: %s\n",
		tga_error_string(tga_get_last_error()));
	width = height = 0;
	return NULL;
    }
    temp_image = new TargaImage(width, height, temp_data);
    free(temp_data);

    result = temp_image->Reverse_Rows();

    delete temp_image;

    return result;
}


static void
RGBA_To_RGB(unsigned char *data, unsigned char *rgb)
{
    unsigned char  alpha = data[3];

    if ( alpha == 0 )
    {
	rgb[0] = BACKGROUND[0];
	rgb[1] = BACKGROUND[1];
	rgb[2] = BACKGROUND[2];
    }
    else
    {
	float	alpha_scale = (float)255 / (float)alpha;
	int	val;
	int	i;

	for ( i = 0 ; i < 3 ; i++ )
	{
	    val = (int)floor(data[i] * alpha_scale);
	    if ( val < 0 )
		rgb[i] = 0;
	    else if ( val > 255 )
		rgb[i] = 255;
	    else
		rgb[i] = val;
	}
    }
}


// Converts an image to RGB form, and returns the rgb pixel data - 24 bits
// per pixel. The returned space should be deleted when no longer required.
unsigned char*
TargaImage::To_RGB(void)
{
    unsigned char   *rgb = new unsigned char[width * height * 3];
    int		    i, j;

    if ( ! data )
	return NULL;

    // Divide out the alpha
    for ( i = 0 ; i < height ; i++ )
    {
	int in_offset = i * width * 4;
	int out_offset = i * width * 3;

	for ( j = 0 ; j < width ; j++ )
        {
	    RGBA_To_RGB(data + (in_offset + j*4), rgb + (out_offset + j*3));
	}
    }

    return rgb;
}

